@@ -106,6 +106,12 @@ static char *str_state[] = {
 	"HEADSET_STATE_PLAYING",
 };
 
+struct headset_state_callback {
+	headset_state_cb cb;
+	void *user_data;
+	unsigned int id;
+};
+
 struct connect_cb {
 	unsigned int id;
 	headset_stream_cb_t cb;
@@ -131,6 +137,7 @@ struct headset {
 	GIOChannel *tmp_rfcomm;
 	GIOChannel *sco;
 	guint sco_id;
+	guint dc_id;
 
 	gboolean auto_dc;
 
@@ -164,6 +171,8 @@ struct event {
 	int (*callback) (struct audio_device *device, const char *buf);
 };
 
+static GSList *headset_callbacks = NULL;
+
 static inline DBusMessage *invalid_args(DBusMessage *msg)
 {
 	return g_dbus_create_error(msg, ERROR_INTERFACE ".InvalidArguments",
@@ -262,10 +271,27 @@ static void print_hf_features(uint32_t features)
 	g_free(str);
 }
 
+static const char *state2str(headset_state_t state)
+{
+	switch (state) {
+	case HEADSET_STATE_DISCONNECTED:
+		return "disconnected";
+	case HEADSET_STATE_CONNECT_IN_PROGRESS:
+		return "connecting";
+	case HEADSET_STATE_CONNECTED:
+	case HEADSET_STATE_PLAY_IN_PROGRESS:
+		return "connected";
+	case HEADSET_STATE_PLAYING:
+		return "playing";
+	}
+
+	return NULL;
+}
+
 static int headset_send_valist(struct headset *hs, char *format, va_list ap)
 {
 	char rsp[BUF_SIZE];
-	ssize_t total_written, written, count;
+	ssize_t total_written, count;
 	int fd;
 
 	count = vsnprintf(rsp, sizeof(rsp), format, ap);
@@ -278,11 +304,14 @@ static int headset_send_valist(struct headset *hs, char *format, va_list ap)
 		return -EIO;
 	}
 
-	written = total_written = 0;
+	total_written = 0;
 	fd = g_io_channel_unix_get_fd(hs->rfcomm);
 
 	while (total_written < count) {
-		written = write(fd, rsp + total_written, count - total_written);
+		ssize_t written;
+
+		written = write(fd, rsp + total_written,
+				count - total_written);
 		if (written < 0)
 			return -errno;
 
@@ -374,6 +403,11 @@ static int report_indicators(struct audio_device *device, const char *buf)
 	if (strlen(buf) < 8)
 		return -EINVAL;
 
+	if (ag.indicators == NULL) {
+		error("HFP AG indicators not initialized");
+		return headset_send(hs, "\r\nERROR\r\n");
+	}
+
 	if (buf[7] == '=')
 		str = indicator_ranges(ag.indicators);
 	else
@@ -452,6 +486,9 @@ static unsigned int connect_cb_new(struct headset *hs,
 
 	pending_connect_init(hs, target_state);
 
+	if (!func)
+		return 0;
+
 	cb = g_new(struct connect_cb, 1);
 
 	cb->cb = func;
@@ -541,7 +578,6 @@ static void sco_connect_cb(GIOChannel *chan, GError *err, gpointer user_data)
 	sk = g_io_channel_unix_get_fd(chan);
 
 	debug("SCO fd=%d", sk);
-	hs->sco = g_io_channel_ref(chan);
 
 	if (p) {
 		p->io = NULL;
@@ -587,7 +623,7 @@ static int sco_connect(struct audio_device *dev, headset_stream_cb_t cb,
 		return -EIO;
 	}
 
-	g_io_channel_unref(io);
+	hs->sco = io;
 
 	headset_set_state(dev, HEADSET_STATE_PLAY_IN_PROGRESS);
 
@@ -1048,7 +1084,8 @@ int telephony_operator_selection_ind(int mode, const char *oper)
 	if (!active_devices)
 		return -ENODEV;
 
-	send_foreach_headset(active_devices, hfp_cmp, "\r\n+COPS: %d,0,%s\r\n",
+	send_foreach_headset(active_devices, hfp_cmp,
+				"\r\n+COPS: %d,0,\"%s\"\r\n",
 				mode, oper);
 	return 0;
 }
@@ -1132,12 +1169,15 @@ static void close_sco(struct audio_device *device)
 	struct headset *hs = device->headset;
 
 	if (hs->sco) {
-		g_source_remove(hs->sco_id);
-		hs->sco_id = 0;
 		g_io_channel_shutdown(hs->sco, TRUE, NULL);
 		g_io_channel_unref(hs->sco);
 		hs->sco = NULL;
 	}
+
+	if (hs->sco_id) {
+		g_source_remove(hs->sco_id);
+		hs->sco_id = 0;
+	}
 }
 
 static gboolean rfcomm_io_cb(GIOChannel *chan, GIOCondition cond,
@@ -1153,8 +1193,10 @@ static gboolean rfcomm_io_cb(GIOChannel *chan, GIOCondition cond,
 
 	hs = device->headset;
 
-	if (cond & (G_IO_ERR | G_IO_HUP))
+	if (cond & (G_IO_ERR | G_IO_HUP)) {
+		debug("ERR or HUP on RFCOMM socket");
 		goto failed;
+	}
 
 	if (g_io_channel_read(chan, (gchar *) buf, sizeof(buf) - 1,
 				&bytes_read) != G_IO_ERROR_NONE)
@@ -1221,13 +1263,9 @@ failed:
 static gboolean sco_cb(GIOChannel *chan, GIOCondition cond,
 			struct audio_device *device)
 {
-	struct headset *hs;
-
 	if (cond & G_IO_NVAL)
 		return FALSE;
 
-	hs = device->headset;
-
 	error("Audio connection got disconnected");
 
 	headset_set_state(device, HEADSET_STATE_CONNECTED);
@@ -1267,12 +1305,6 @@ void headset_connect_cb(GIOChannel *chan, GError *err, gpointer user_data)
 	else
 		hs->auto_dc = FALSE;
 
-	if (server_is_enabled(&dev->src, HANDSFREE_SVCLASS_ID) &&
-			hs->hfp_handle != 0)
-		hs->hfp_active = TRUE;
-	else
-		hs->hfp_active = FALSE;
-
 	g_io_add_watch(chan, G_IO_IN | G_IO_ERR | G_IO_HUP| G_IO_NVAL,
 			(GIOFunc) rfcomm_io_cb, dev);
 
@@ -1310,14 +1342,38 @@ failed:
 		headset_set_state(dev, HEADSET_STATE_DISCONNECTED);
 }
 
+static void headset_set_channel(struct headset *headset,
+				const sdp_record_t *record, uint16_t svc)
+{
+	int ch;
+	sdp_list_t *protos;
+
+	if (sdp_get_access_protos(record, &protos) < 0) {
+		error("Unable to get access protos from headset record");
+		return;
+	}
+
+	ch = sdp_get_proto_port(protos, RFCOMM_UUID);
+
+	sdp_list_foreach(protos, (sdp_list_func_t) sdp_list_free, NULL);
+	sdp_list_free(protos, NULL);
+
+	if (ch > 0) {
+		headset->rfcomm_ch = ch;
+		debug("Discovered %s service on RFCOMM channel %d",
+			svc == HEADSET_SVCLASS_ID ? "Headset" : "Handsfree",
+			ch);
+	} else
+		error("Unable to get RFCOMM channel from Headset record");
+}
+
 static void get_record_cb(sdp_list_t *recs, int err, gpointer user_data)
 {
 	struct audio_device *dev = user_data;
 	struct headset *hs = dev->headset;
 	struct pending_connect *p = hs->pending;
-	int ch = -1;
 	sdp_record_t *record = NULL;
-	sdp_list_t *protos, *classes = NULL;
+	sdp_list_t *classes = NULL;
 	uuid_t uuid;
 
 	if (err < 0) {
@@ -1359,24 +1415,16 @@ static void get_record_cb(sdp_list_t *recs, int err, gpointer user_data)
 		hs->hsp_handle = record->handle;
 	}
 
-	if (!sdp_get_access_protos(record, &protos)) {
-		ch = sdp_get_proto_port(protos, RFCOMM_UUID);
-		sdp_list_foreach(protos, (sdp_list_func_t) sdp_list_free,
-					NULL);
-		sdp_list_free(protos, NULL);
-		protos = NULL;
-	}
+	headset_set_channel(hs, record, uuid.value.uuid16);
 
-	if (ch == -1) {
+	if (hs->rfcomm_ch == -1) {
 		error("Unable to extract RFCOMM channel from service record");
 		goto failed_not_supported;
 	}
 
-	hs->rfcomm_ch = ch;
-
 	err = rfcomm_connect(dev, NULL, NULL, NULL);
 	if (err < 0) {
-		error("Unable to connect: %s (%s)", strerror(-err), -err);
+		error("Unable to connect: %s (%d)", strerror(-err), -err);
 		p->err = -err;
 		error_connection_attempt_failed(dev->conn, p->msg, p->err);
 		goto failed;
@@ -1443,12 +1491,17 @@ static int rfcomm_connect(struct audio_device *dev, headset_stream_cb_t cb,
 				BT_IO_OPT_DEST_BDADDR, &dev->dst,
 				BT_IO_OPT_CHANNEL, hs->rfcomm_ch,
 				BT_IO_OPT_INVALID);
+
+	hs->rfcomm_ch = -1;
+
 	if (!io) {
 		error("%s", err->message);
 		g_error_free(err);
 		return -EIO;
 	}
 
+	hs->hfp_active = hs->hfp_handle != 0 ? TRUE : FALSE;
+
 	g_io_channel_unref(io);
 
 	headset_set_state(dev, HEADSET_STATE_CONNECT_IN_PROGRESS);
@@ -1511,13 +1564,8 @@ static DBusMessage *hs_disconnect(DBusConnection *conn, DBusMessage *msg,
 {
 	struct audio_device *device = data;
 	struct headset *hs = device->headset;
-	DBusMessage *reply = NULL;
 	char hs_address[18];
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return NULL;
-
 	if (hs->state == HEADSET_STATE_DISCONNECTED)
 		return g_dbus_create_error(msg, ERROR_INTERFACE
 						".NotConnected",
@@ -1527,7 +1575,8 @@ static DBusMessage *hs_disconnect(DBusConnection *conn, DBusMessage *msg,
 	ba2str(&device->dst, hs_address);
 	info("Disconnected from %s, %s", hs_address, device->path);
 
-	return reply;
+	return dbus_message_new_method_return(msg);
+
 }
 
 static DBusMessage *hs_is_connected(DBusConnection *conn,
@@ -1573,6 +1622,8 @@ static DBusMessage *hs_connect(DBusConnection *conn, DBusMessage *msg,
 		return g_dbus_create_error(msg, ERROR_INTERFACE ".NotAllowed",
 						"Too many connected devices");
 
+	device->auto_connect = FALSE;
+
 	err = rfcomm_connect(device, NULL, NULL, NULL);
 	if (err < 0)
 		return g_dbus_create_error(msg, ERROR_INTERFACE
@@ -1834,6 +1885,7 @@ static DBusMessage *hs_get_properties(DBusConnection *conn,
 	DBusMessageIter iter;
 	DBusMessageIter dict;
 	gboolean value;
+	const char *state;
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
@@ -1851,6 +1903,11 @@ static DBusMessage *hs_get_properties(DBusConnection *conn,
 	value = (device->headset->state == HEADSET_STATE_PLAYING);
 	dict_append_entry(&dict, "Playing", DBUS_TYPE_BOOLEAN, &value);
 
+	/* State */
+	state = state2str(device->headset->state);
+	if (state)
+		dict_append_entry(&dict, "State", DBUS_TYPE_STRING, &state);
+
 	/* Connected */
 	value = (device->headset->state >= HEADSET_STATE_CONNECTED);
 	dict_append_entry(&dict, "Connected", DBUS_TYPE_BOOLEAN, &value);
@@ -1949,31 +2006,6 @@ static GDBusSignalTable headset_signals[] = {
 	{ NULL, NULL }
 };
 
-static void headset_set_channel(struct headset *headset,
-				const sdp_record_t *record, uint16_t svc)
-{
-	int ch;
-	sdp_list_t *protos;
-
-	if (sdp_get_access_protos(record, &protos) < 0) {
-		error("Unable to get access protos from headset record");
-		return;
-	}
-
-	ch = sdp_get_proto_port(protos, RFCOMM_UUID);
-
-	sdp_list_foreach(protos, (sdp_list_func_t) sdp_list_free, NULL);
-	sdp_list_free(protos, NULL);
-
-	if (ch > 0) {
-		headset->rfcomm_ch = ch;
-		debug("Discovered %s service on RFCOMM channel %d",
-			svc == HEADSET_SVCLASS_ID ? "Headset" : "Handsfree",
-			ch);
-	} else
-		error("Unable to get RFCOMM channel from Headset record");
-}
-
 void headset_update(struct audio_device *dev, uint16_t svc,
 			const char *uuidstr)
 {
@@ -2014,8 +2046,6 @@ void headset_update(struct audio_device *dev, uint16_t svc,
 		debug("Invalid record passed to headset_update");
 		return;
 	}
-
-	headset_set_channel(headset, record, svc);
 }
 
 static void headset_free(struct audio_device *dev)
@@ -2027,6 +2057,9 @@ static void headset_free(struct audio_device *dev)
 		hs->dc_timer = 0;
 	}
 
+	if (hs->dc_id)
+		device_remove_disconnect_watch(dev->btd_dev, hs->dc_id);
+
 	if (hs->sco) {
 		g_io_channel_shutdown(hs->sco, TRUE, NULL);
 		g_io_channel_unref(hs->sco);
@@ -2048,10 +2081,11 @@ static void path_unregister(void *data)
 
 	if (hs->state > HEADSET_STATE_DISCONNECTED) {
 		debug("Headset unregistered while device was connected!");
+		bt_cancel_discovery(&dev->src, &dev->dst);
 		headset_set_state(dev, HEADSET_STATE_DISCONNECTED);
 	}
 
-	info("Unregistered interface %s on path %s",
+	debug("Unregistered interface %s on path %s",
 		AUDIO_HEADSET_INTERFACE, dev->path);
 
 	headset_free(dev);
@@ -2097,7 +2131,6 @@ struct headset *headset_init(struct audio_device *dev, uint16_t svc,
 		return NULL;
 	}
 
-	headset_set_channel(hs, record, svc);
 register_iface:
 	if (!g_dbus_register_interface(dev->conn, dev->path,
 					AUDIO_HEADSET_INTERFACE,
@@ -2107,7 +2140,7 @@ register_iface:
 		return NULL;
 	}
 
-	info("Registered interface %s on path %s",
+	debug("Registered interface %s on path %s",
 		AUDIO_HEADSET_INTERFACE, dev->path);
 
 	return hs;
@@ -2194,13 +2227,12 @@ static gboolean dummy_connect_complete(struct audio_device *dev)
 
 unsigned int headset_request_stream(struct audio_device *dev,
 					headset_stream_cb_t cb,
-					headset_lock_t lock,
 					void *user_data)
 {
 	struct headset *hs = dev->headset;
 	unsigned int id;
 
-	if (hs->rfcomm && hs->sco) {
+	if (hs->state == HEADSET_STATE_PLAYING) {
 		id = connect_cb_new(hs, HEADSET_STATE_PLAYING, cb, user_data);
 		g_idle_add((GSourceFunc) dummy_connect_complete, dev);
 		return id;
@@ -2219,7 +2251,7 @@ unsigned int headset_request_stream(struct audio_device *dev,
 		if (rfcomm_connect(dev, cb, user_data, &id) < 0)
 			return 0;
 		hs->auto_dc = TRUE;
-	} else {
+	} else if (hs->sco == NULL) {
 		if (sco_connect(dev, cb, user_data, &id) < 0)
 			return 0;
 	}
@@ -2230,15 +2262,12 @@ unsigned int headset_request_stream(struct audio_device *dev,
 }
 
 unsigned int headset_config_stream(struct audio_device *dev,
+					gboolean auto_dc,
 					headset_stream_cb_t cb,
-					headset_lock_t lock,
 					void *user_data)
 {
 	struct headset *hs = dev->headset;
-	unsigned int id;
-
-	if (hs->lock & lock)
-		return 0;
+	unsigned int id = 0;
 
 	if (hs->dc_timer) {
 		g_source_remove(hs->dc_timer);
@@ -2255,7 +2284,7 @@ unsigned int headset_config_stream(struct audio_device *dev,
 	if (rfcomm_connect(dev, cb, user_data, &id) < 0)
 		return 0;
 
-	hs->auto_dc = TRUE;
+	hs->auto_dc = auto_dc;
 	hs->pending->target_state = HEADSET_STATE_CONNECTED;
 
 	return id;
@@ -2268,15 +2297,11 @@ done:
 
 unsigned int headset_suspend_stream(struct audio_device *dev,
 					headset_stream_cb_t cb,
-					headset_lock_t lock,
 					void *user_data)
 {
 	struct headset *hs = dev->headset;
 	unsigned int id;
 
-	if (hs->lock & ~lock || !hs->rfcomm || !hs->sco)
-		return 0;
-
 	if (hs->dc_timer) {
 		g_source_remove(hs->dc_timer);
 		hs->dc_timer = 0;
@@ -2363,33 +2388,62 @@ static int headset_close_rfcomm(struct audio_device *dev)
 	return 0;
 }
 
+static void disconnect_cb(struct btd_device *btd_dev, gboolean removal,
+				void *user_data)
+{
+	struct audio_device *device = user_data;
+
+	info("Headset: disconnect %s", device->path);
+
+	headset_set_state(device, HEADSET_STATE_DISCONNECTED);
+}
+
 void headset_set_state(struct audio_device *dev, headset_state_t state)
 {
 	struct headset *hs = dev->headset;
 	gboolean value;
+	const char *state_str;
+	headset_state_t old_state = hs->state;
+	GSList *l;
 
-	if (hs->state == state)
+	if (old_state == state)
 		return;
 
+	state_str = state2str(state);
+
 	switch (state) {
 	case HEADSET_STATE_DISCONNECTED:
 		value = FALSE;
 		close_sco(dev);
 		headset_close_rfcomm(dev);
+		emit_property_changed(dev->conn, dev->path,
+					AUDIO_HEADSET_INTERFACE, "State",
+					DBUS_TYPE_STRING, &state_str);
 		g_dbus_emit_signal(dev->conn, dev->path,
 					AUDIO_HEADSET_INTERFACE,
 					"Disconnected",
 					DBUS_TYPE_INVALID);
-		emit_property_changed(dev->conn, dev->path,
+		if (hs->state > HEADSET_STATE_CONNECT_IN_PROGRESS)
+			emit_property_changed(dev->conn, dev->path,
 					AUDIO_HEADSET_INTERFACE, "Connected",
 					DBUS_TYPE_BOOLEAN, &value);
+
 		telephony_device_disconnected(dev);
 		active_devices = g_slist_remove(active_devices, dev);
+		device_remove_disconnect_watch(dev->btd_dev, hs->dc_id);
+		hs->dc_id = 0;
 		break;
 	case HEADSET_STATE_CONNECT_IN_PROGRESS:
+		emit_property_changed(dev->conn, dev->path,
+					AUDIO_HEADSET_INTERFACE, "State",
+					DBUS_TYPE_STRING, &state_str);
 		break;
 	case HEADSET_STATE_CONNECTED:
 		close_sco(dev);
+		if (hs->state != HEADSET_STATE_PLAY_IN_PROGRESS)
+			emit_property_changed(dev->conn, dev->path,
+					AUDIO_HEADSET_INTERFACE, "State",
+					DBUS_TYPE_STRING, &state_str);
 		if (hs->state < state) {
 			value = TRUE;
 			g_dbus_emit_signal(dev->conn, dev->path,
@@ -2402,6 +2456,9 @@ void headset_set_state(struct audio_device *dev, headset_state_t state)
 						DBUS_TYPE_BOOLEAN, &value);
 			active_devices = g_slist_append(active_devices, dev);
 			telephony_device_connected(dev);
+			hs->dc_id = device_add_disconnect_watch(dev->btd_dev,
+								disconnect_cb,
+								dev, NULL);
 		} else if (hs->state == HEADSET_STATE_PLAYING) {
 			value = FALSE;
 			g_dbus_emit_signal(dev->conn, dev->path,
@@ -2418,6 +2475,9 @@ void headset_set_state(struct audio_device *dev, headset_state_t state)
 		break;
 	case HEADSET_STATE_PLAYING:
 		value = TRUE;
+		emit_property_changed(dev->conn, dev->path,
+					AUDIO_HEADSET_INTERFACE, "State",
+					DBUS_TYPE_STRING, &state_str);
 		hs->sco_id = g_io_add_watch(hs->sco,
 					G_IO_ERR | G_IO_HUP | G_IO_NVAL,
 					(GIOFunc) sco_cb, dev);
@@ -2436,9 +2496,15 @@ void headset_set_state(struct audio_device *dev, headset_state_t state)
 		break;
 	}
 
-	debug("State changed %s: %s -> %s", dev->path, str_state[hs->state],
-		str_state[state]);
 	hs->state = state;
+
+	debug("State changed %s: %s -> %s", dev->path, str_state[old_state],
+		str_state[state]);
+
+	for (l = headset_callbacks; l != NULL; l = l->next) {
+		struct headset_state_callback *cb = l->data;
+		cb->cb(dev, old_state, state, cb->user_data);
+	}
 }
 
 headset_state_t headset_get_state(struct audio_device *dev)
@@ -2465,6 +2531,13 @@ gboolean headset_is_active(struct audio_device *dev)
 	return FALSE;
 }
 
+headset_lock_t headset_get_lock(struct audio_device *dev)
+{
+	struct headset *hs = dev->headset;
+
+	return hs->lock;
+}
+
 gboolean headset_lock(struct audio_device *dev, headset_lock_t lock)
 {
 	struct headset *hs = dev->headset;
@@ -2587,22 +2660,18 @@ int telephony_incoming_call_ind(const char *number, int type)
 		return -EBUSY;
 	}
 
+	/* With HSP 1.2 the RING messages should *not* be sent if inband
+	 * ringtone is being used */
+	if (!hs->hfp_active && (ag.features & AG_FEATURE_INBAND_RINGTONE))
+		return 0;
+
 	g_free(ag.number);
 	ag.number = g_strdup(number);
 	ag.number_type = type;
 
 	if (ag.features & AG_FEATURE_INBAND_RINGTONE && hs->hfp_active &&
 			hs->state != HEADSET_STATE_PLAYING) {
-		if (hs->state == HEADSET_STATE_CONNECTED) {
-			int ret;
-
-			ret = sco_connect(dev, NULL, NULL, NULL);
-			if (ret < 0)
-				return ret;
-		}
-
 		hs->pending_ring = TRUE;
-
 		return 0;
 	}
 
@@ -2662,9 +2731,8 @@ int telephony_list_current_call_ind(int idx, int dir, int status, int mode,
 
 	if (number)
 		send_foreach_headset(active_devices, hfp_cmp,
-					"\r\n+CLCC: %d,%d,%d,%d,%d,%s,%d\r\n",
-					idx, dir, status, mode, mprty,
-					number, type);
+				"\r\n+CLCC: %d,%d,%d,%d,%d,\"%s\",%d\r\n",
+				idx, dir, status, mode, mprty, number, type);
 	else
 		send_foreach_headset(active_devices, hfp_cmp,
 					"\r\n+CLCC: %d,%d,%d,%d,%d\r\n",
@@ -2706,3 +2774,34 @@ int telephony_call_waiting_ind(const char *number, int type)
 
 	return 0;
 }
+
+unsigned int headset_add_state_cb(headset_state_cb cb, void *user_data)
+{
+	struct headset_state_callback *state_cb;
+	static unsigned int id = 0;
+
+	state_cb = g_new(struct headset_state_callback, 1);
+	state_cb->cb = cb;
+	state_cb->user_data = user_data;
+	state_cb->id = ++id;
+
+	headset_callbacks = g_slist_append(headset_callbacks, state_cb);
+
+	return state_cb->id;
+}
+
+gboolean headset_remove_state_cb(unsigned int id)
+{
+	GSList *l;
+
+	for (l = headset_callbacks; l != NULL; l = l->next) {
+		struct headset_state_callback *cb = l->data;
+		if (cb && cb->id == id) {
+			headset_callbacks = g_slist_remove(headset_callbacks, cb);
+			g_free(cb);
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
