@@ -3,7 +3,7 @@
  *  BlueZ - Bluetooth protocol stack for Linux
  *
  *  Copyright (C) 2006-2007  Nokia Corporation
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This library is free software; you can redistribute it and/or
@@ -96,10 +96,10 @@ struct bluetooth_a2dp {
 	sbc_capabilities_t sbc_capabilities;
 	sbc_t sbc;				/* Codec data */
 	int sbc_initialized;			/* Keep track if the encoder is initialized */
-	int codesize;				/* SBC codesize */
+	unsigned int codesize;			/* SBC codesize */
 	int samples;				/* Number of encoded samples */
 	uint8_t buffer[BUFFER_SIZE];		/* Codec transfer buffer */
-	int count;				/* Codec transfer buffer counter */
+	unsigned int count;			/* Codec transfer buffer counter */
 
 	int nsamples;				/* Cumulative number of codec samples */
 	uint16_t seq_num;			/* Cumulative packet sequence */
@@ -131,11 +131,11 @@ struct bluetooth_data {
 	struct bluetooth_alsa_config alsa_config;	/* ALSA resource file parameters */
 	volatile snd_pcm_sframes_t hw_ptr;
 	int transport;					/* chosen transport SCO or AD2P */
-	int link_mtu;					/* MTU for selected transport channel */
+	unsigned int link_mtu;				/* MTU for selected transport channel */
 	volatile struct pollfd stream;			/* Audio stream filedescriptor */
 	struct pollfd server;				/* Audio daemon filedescriptor */
 	uint8_t buffer[BUFFER_SIZE];		/* Encoded transfer buffer */
-	int count;					/* Transfer buffer counter */
+	unsigned int count;				/* Transfer buffer counter */
 	struct bluetooth_a2dp a2dp;			/* A2DP data */
 
 	pthread_t hw_thread;				/* Makes virtual hw pointer move */
@@ -146,7 +146,7 @@ struct bluetooth_data {
 
 static int audioservice_send(int sk, const bt_audio_msg_header_t *msg);
 static int audioservice_expect(int sk, bt_audio_msg_header_t *outmsg,
-				int expected_type);
+							int expected_type);
 
 static int bluetooth_start(snd_pcm_ioplug_t *io)
 {
@@ -224,7 +224,7 @@ static void *playback_hw_thread(void *param)
 
 			/* Reset point of reference to avoid too big values
 			 * that wont fit an unsigned int */
-			if (delta.tv_sec < UINT_SECS_MAX)
+			if ((unsigned int) delta.tv_sec < UINT_SECS_MAX)
 				prev_periods = periods;
 			else {
 				prev_periods = 0;
@@ -430,6 +430,8 @@ static int bluetooth_hsp_hw_params(snd_pcm_ioplug_t *io,
 {
 	struct bluetooth_data *data = io->private_data;
 	char buf[BT_SUGGESTED_BUFFER_SIZE];
+	struct bt_open_req *open_req = (void *) buf;
+	struct bt_open_rsp *open_rsp = (void *) buf;
 	struct bt_set_configuration_req *req = (void *) buf;
 	struct bt_set_configuration_rsp *rsp = (void *) buf;
 	int err;
@@ -437,17 +439,34 @@ static int bluetooth_hsp_hw_params(snd_pcm_ioplug_t *io,
 	DBG("Preparing with io->period_size=%lu io->buffer_size=%lu",
 					io->period_size, io->buffer_size);
 
+	memset(req, 0, BT_SUGGESTED_BUFFER_SIZE);
+	open_req->h.type = BT_REQUEST;
+	open_req->h.name = BT_OPEN;
+	open_req->h.length = sizeof(*open_req);
+
+	strncpy(open_req->destination, data->alsa_config.device, 18);
+	open_req->seid = BT_A2DP_SEID_RANGE + 1;
+	open_req->lock = (io->stream == SND_PCM_STREAM_PLAYBACK ?
+			BT_WRITE_LOCK : BT_READ_LOCK);
+
+	err = audioservice_send(data->server.fd, &open_req->h);
+	if (err < 0)
+		return err;
+
+	open_rsp->h.length = sizeof(*open_rsp);
+	err = audioservice_expect(data->server.fd, &open_rsp->h,
+					BT_OPEN);
+	if (err < 0)
+		return err;
+
 	memset(req, 0, BT_SUGGESTED_BUFFER_SIZE);
 	req->h.type = BT_REQUEST;
 	req->h.name = BT_SET_CONFIGURATION;
 	req->h.length = sizeof(*req);
 
-	strncpy(req->device, data->alsa_config.device, 18);
 	req->codec.transport = BT_CAPABILITIES_TRANSPORT_SCO;
+	req->codec.seid = BT_A2DP_SEID_RANGE + 1;
 	req->codec.length = sizeof(pcm_capabilities_t);
-	req->access_mode = (io->stream == SND_PCM_STREAM_PLAYBACK ?
-			BT_CAPABILITIES_ACCESS_MODE_WRITE :
-			BT_CAPABILITIES_ACCESS_MODE_READ);
 
 	req->h.length += req->codec.length - sizeof(req->codec);
 	err = audioservice_send(data->server.fd, &req->h);
@@ -460,7 +479,7 @@ static int bluetooth_hsp_hw_params(snd_pcm_ioplug_t *io,
 	if (err < 0)
 		return err;
 
-	data->transport = rsp->transport;
+	data->transport = BT_CAPABILITIES_TRANSPORT_SCO;
 	data->link_mtu = rsp->link_mtu;
 
 	return 0;
@@ -503,7 +522,7 @@ static uint8_t default_bitpool(uint8_t freq, uint8_t mode)
 }
 
 static int bluetooth_a2dp_init(struct bluetooth_data *data,
-				snd_pcm_hw_params_t *params)
+					snd_pcm_hw_params_t *params)
 {
 	struct bluetooth_alsa_config *cfg = &data->alsa_config;
 	sbc_capabilities_t *cap = &data->a2dp.sbc_capabilities;
@@ -671,6 +690,8 @@ static int bluetooth_a2dp_hw_params(snd_pcm_ioplug_t *io,
 	struct bluetooth_data *data = io->private_data;
 	struct bluetooth_a2dp *a2dp = &data->a2dp;
 	char buf[BT_SUGGESTED_BUFFER_SIZE];
+	struct bt_open_req *open_req = (void *) buf;
+	struct bt_open_rsp *open_rsp = (void *) buf;
 	struct bt_set_configuration_req *req = (void *) buf;
 	struct bt_set_configuration_rsp *rsp = (void *) buf;
 	int err;
@@ -678,6 +699,26 @@ static int bluetooth_a2dp_hw_params(snd_pcm_ioplug_t *io,
 	DBG("Preparing with io->period_size=%lu io->buffer_size=%lu",
 					io->period_size, io->buffer_size);
 
+	memset(req, 0, BT_SUGGESTED_BUFFER_SIZE);
+	open_req->h.type = BT_REQUEST;
+	open_req->h.name = BT_OPEN;
+	open_req->h.length = sizeof(*open_req);
+
+	strncpy(open_req->destination, data->alsa_config.device, 18);
+	open_req->seid = a2dp->sbc_capabilities.capability.seid;
+	open_req->lock = (io->stream == SND_PCM_STREAM_PLAYBACK ?
+			BT_WRITE_LOCK : BT_READ_LOCK);
+
+	err = audioservice_send(data->server.fd, &open_req->h);
+	if (err < 0)
+		return err;
+
+	open_rsp->h.length = sizeof(*open_rsp);
+	err = audioservice_expect(data->server.fd, &open_rsp->h,
+					BT_OPEN);
+	if (err < 0)
+		return err;
+
 	err = bluetooth_a2dp_init(data, params);
 	if (err < 0)
 		return err;
@@ -687,15 +728,11 @@ static int bluetooth_a2dp_hw_params(snd_pcm_ioplug_t *io,
 	req->h.name = BT_SET_CONFIGURATION;
 	req->h.length = sizeof(*req);
 
-	strncpy(req->device, data->alsa_config.device, 18);
 	memcpy(&req->codec, &a2dp->sbc_capabilities,
 			sizeof(a2dp->sbc_capabilities));
 
 	req->codec.transport = BT_CAPABILITIES_TRANSPORT_A2DP;
 	req->codec.length = sizeof(a2dp->sbc_capabilities);
-	req->access_mode = (io->stream == SND_PCM_STREAM_PLAYBACK ?
-			BT_CAPABILITIES_ACCESS_MODE_WRITE :
-			BT_CAPABILITIES_ACCESS_MODE_READ);
 	req->h.length += req->codec.length - sizeof(req->codec);
 
 	err = audioservice_send(data->server.fd, &req->h);
@@ -708,7 +745,7 @@ static int bluetooth_a2dp_hw_params(snd_pcm_ioplug_t *io,
 	if (err < 0)
 		return err;
 
-	data->transport = rsp->transport;
+	data->transport = BT_CAPABILITIES_TRANSPORT_A2DP;
 	data->link_mtu = rsp->link_mtu;
 
 	/* Setup SBC encoder now we agree on parameters */
@@ -797,8 +834,7 @@ static int bluetooth_playback_poll_revents(snd_pcm_ioplug_t *io,
 	if (pfds[1].revents & (POLLERR | POLLHUP | POLLNVAL))
 		io->state = SND_PCM_STATE_DISCONNECTED;
 
-	revents[0] = (pfds[0].revents & ~POLLIN) | POLLOUT;
-	revents[1] = (pfds[1].revents & ~POLLIN);
+	*revents = (pfds[0].revents & POLLIN) ? POLLOUT : 0;
 
 	return 0;
 }
@@ -812,7 +848,8 @@ static snd_pcm_sframes_t bluetooth_hsp_read(snd_pcm_ioplug_t *io,
 	struct bluetooth_data *data = io->private_data;
 	snd_pcm_uframes_t frames_to_write, ret;
 	unsigned char *buff;
-	int nrecv, frame_size = 0;
+	unsigned int frame_size = 0;
+	int nrecv;
 
 	DBG("areas->step=%u areas->first=%u offset=%lu size=%lu io->nonblock=%u",
 			areas->step, areas->first, offset, size, io->nonblock);
@@ -830,7 +867,7 @@ static snd_pcm_sframes_t bluetooth_hsp_read(snd_pcm_ioplug_t *io,
 		goto done;
 	}
 
-	if (nrecv != data->link_mtu) {
+	if ((unsigned int) nrecv != data->link_mtu) {
 		ret = -EIO;
 		SNDERR(strerror(-ret));
 		goto done;
@@ -921,8 +958,7 @@ done:
 
 static snd_pcm_sframes_t bluetooth_a2dp_read(snd_pcm_ioplug_t *io,
 				const snd_pcm_channel_area_t *areas,
-				snd_pcm_uframes_t offset,
-				snd_pcm_uframes_t size)
+				snd_pcm_uframes_t offset, snd_pcm_uframes_t size)
 {
 	snd_pcm_uframes_t ret = 0;
 	return ret;
@@ -969,7 +1005,9 @@ static snd_pcm_sframes_t bluetooth_a2dp_write(snd_pcm_ioplug_t *io,
 	struct bluetooth_data *data = io->private_data;
 	struct bluetooth_a2dp *a2dp = &data->a2dp;
 	snd_pcm_sframes_t ret = 0;
-	int frame_size, encoded, written, bytes_left;
+	unsigned int bytes_left;
+	int frame_size, encoded;
+	size_t written;
 	uint8_t *buff;
 
 	DBG("areas->step=%u areas->first=%u offset=%lu size=%lu",
@@ -1182,7 +1220,7 @@ static int bluetooth_hsp_hw_constraint(snd_pcm_ioplug_t *io)
 		SND_PCM_ACCESS_MMAP_INTERLEAVED
 	};
 	unsigned int format_list[] = {
-		SND_PCM_FORMAT_S16_LE
+		SND_PCM_FORMAT_S16
 	};
 	int err;
 
@@ -1237,7 +1275,7 @@ static int bluetooth_a2dp_hw_constraint(snd_pcm_ioplug_t *io)
 		SND_PCM_ACCESS_MMAP_INTERLEAVED
 	};
 	unsigned int format_list[] = {
-		SND_PCM_FORMAT_S16_LE
+		SND_PCM_FORMAT_S16
 	};
 	unsigned int rate_list[4];
 	unsigned int rate_count;
@@ -1495,7 +1533,8 @@ static int audioservice_send(int sk, const bt_audio_msg_header_t *msg)
 
 	length = msg->length ? msg->length : BT_SUGGESTED_BUFFER_SIZE;
 
-	DBG("sending %s", bt_audio_strmsg(msg->msg_type));
+	DBG("sending %s:%s", bt_audio_strtype(msg->type),
+		bt_audio_strname(msg->name));
 	if (send(sk, msg, length, 0) > 0)
 		err = 0;
 	else {
@@ -1510,13 +1549,23 @@ static int audioservice_send(int sk, const bt_audio_msg_header_t *msg)
 static int audioservice_recv(int sk, bt_audio_msg_header_t *inmsg)
 {
 	int err;
+	ssize_t ret;
 	const char *type, *name;
 	uint16_t length;
 
 	length = inmsg->length ? inmsg->length : BT_SUGGESTED_BUFFER_SIZE;
 
 	DBG("trying to receive msg from audio service...");
-	if (recv(sk, inmsg, length, 0) > 0) {
+
+	ret = recv(sk, inmsg, length, 0);
+	if (ret < 0) {
+		err = -errno;
+		SNDERR("Error receiving IPC data from bluetoothd: %s (%d)",
+						strerror(errno), errno);
+	} else if ((size_t) ret < sizeof(bt_audio_msg_header_t)) {
+		SNDERR("Too short (%d bytes) IPC packet from bluetoothd", ret);
+		err = -EINVAL;
+	} else {
 		type = bt_audio_strtype(inmsg->type);
 		name = bt_audio_strname(inmsg->name);
 		if (type && name) {
@@ -1525,20 +1574,17 @@ static int audioservice_recv(int sk, bt_audio_msg_header_t *inmsg)
 		} else {
 			err = -EINVAL;
 			SNDERR("Bogus message type %d - name %d"
-					"received from audio service",
+					" received from audio service",
 					inmsg->type, inmsg->name);
 		}
-	} else {
-		err = -errno;
-		SNDERR("Error receiving data from audio service: %s(%d)",
-					strerror(errno), errno);
+
 	}
 
 	return err;
 }
 
 static int audioservice_expect(int sk, bt_audio_msg_header_t *rsp,
-				int expected_name)
+							int expected_name)
 {
 	bt_audio_error_t *error;
 	int err = audioservice_recv(sk, rsp);
@@ -1577,7 +1623,8 @@ static int bluetooth_parse_capabilities(struct bluetooth_data *data,
 		return 0;
 
 	while (bytes_left > 0) {
-		if (codec->type == BT_A2DP_CODEC_SBC)
+		if ((codec->type == BT_A2DP_SBC_SINK) &&
+				!(codec->lock & BT_WRITE_LOCK))
 			break;
 
 		bytes_left -= codec->length;
@@ -1593,8 +1640,8 @@ static int bluetooth_parse_capabilities(struct bluetooth_data *data,
 	return 0;
 }
 
-static int bluetooth_init(struct bluetooth_data *data, snd_pcm_stream_t stream,
-				snd_config_t *conf)
+static int bluetooth_init(struct bluetooth_data *data,
+				snd_pcm_stream_t stream, snd_config_t *conf)
 {
 	int sk, err;
 	struct bluetooth_alsa_config *alsa_conf = &data->alsa_config;
@@ -1643,7 +1690,7 @@ static int bluetooth_init(struct bluetooth_data *data, snd_pcm_stream_t stream,
 
 	if (alsa_conf->autoconnect)
 		req->flags |= BT_FLAG_AUTOCONNECT;
-	strncpy(req->device, alsa_conf->device, 18);
+	strncpy(req->destination, alsa_conf->device, 18);
 	if (alsa_conf->has_transport)
 		req->transport = alsa_conf->transport;
 	else
@@ -1664,7 +1711,8 @@ static int bluetooth_init(struct bluetooth_data *data, snd_pcm_stream_t stream,
 	return 0;
 
 failed:
-	bt_audio_service_close(sk);
+	if (sk >= 0)
+		bt_audio_service_close(sk);
 	return err;
 }
 
